import org.w3c.dom.*;
import java.io.*;

public class Parser
{
	XMLNode Doc, cNode;
	String cNodeName = "-99N";
	int cNodeDepth = -99;
	boolean cNodeOpen=false;
	InputStreamReader xstream;

	static final short BLK_TAGBEGIN=1;
	static final short BLK_TAGEND=2;
	static final short BLK_TAGNAME=3;
	static final short BLK_ATTR=4;
	static final short BLK_CHRDATA=5;
	static final short BLK_TEXTDATA=6;
	static final short BLK_MISC=7;
	static final short BLK_ENTITY=8;
	static final short BLK_DOCTYPE1=9;
	static final short BLK_DOCTYPE2=10;
	static final short BLK_PI=11;

	static final int xBufSize=256000, dataBufSize=16384;
	int i, charsRead, xdataLen=0, lineCount, charPos=0, equalCount=0;
	char[] xbuf;
	char[] xdata;

	short curBLK = BLK_TAGBEGIN;
	boolean tagOpen=false;

	char[] cdataMarkup = { 'C', 'D', 'A', 'T', 'A', '[' };
	char[] entityMarkup = { 'N', 'T', 'I', 'T', 'Y' };
	char[] doctypeMarkup = { 'O', 'C', 'T', 'Y', 'P', 'E' };
	char[] commentMarkup = { '-' };
	short markupPos  = 0;
	char  markCHAR=' ', quoteCHAR, tagQualifier=' ';

	public Document load(String xmlFile)
	{
		try {
long stime = System.currentTimeMillis();
		FileInputStream xf = new FileInputStream(xmlFile);
System.out.println("Opened File in "+(System.currentTimeMillis()-stime)+" ms");
stime = System.currentTimeMillis();
		//xstream = new InputStreamReader(xf);
		//xbuf  = new char[xBufSize];
		parse(xf);
		/*do  {
		refreshxBuffer();
System.out.println("CharsRead: "+charsRead);
		//for(int n=0; n < charsRead; n++) markCHAR = xbuf[n];
		} while(charsRead > -1);*/
System.out.println("Done Parsing in "+(System.currentTimeMillis()-stime)+" ms");
		xf.close();
		} catch(Exception e) { System.out.println(e.toString()); }
		return null;
	}

	Document parse(InputStream xs)
	{
		int markCount=0;
		char xch;
		boolean iFlag = false, docBegin=false;
		String curTagName;
		char[] cMarkup=null;

		xstream = new InputStreamReader(xs);
		xbuf  = new char[xBufSize];
		xdata = new char[dataBufSize];

		Doc = new XMLNode();
		Doc.nodeType = Node.DOCUMENT_NODE;
		Doc.nodeData = new Object[Xdef.DOCUMENT_DATA_COUNT];

		try
		{

		lineCount = 1;
		refreshxBuffer();
		xch = xbuf[0];

		//--------------------------------------------------------------
		//            ParserLoop - Implements Parser Logic
		//--------------------------------------------------------------

		ParserLoop:
		do
		{

		//Core Loop for parsing Tags, Attributes, Text, CDATA & WhiteSpaces
		CoreLoop:
		do
		{
			//TagName begin : most frequent case
			if(curBLK == BLK_TAGNAME) {
				if((xch > '@' && xch < '[') || (xch > '`' && xch < '{') || xch == ':' || xch == '_')
				{
					xdata[0] = xch;
					xdataLen = 1;
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				}
				else { curBLK = BLK_TAGEND; continue CoreLoop; }

				while((xch > '@' && xch < '[') || (xch > '`' && xch < '{') ||
					(xch > '/' && xch <= '9') || xch == ':' || xch == '_' || xch == '-')
				{
					xdata[xdataLen] = xch;
					xdataLen++;
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				}

				curBLK = handleTAG();
				//let it fall to whitespace eliminator
			}


			else if(curBLK == BLK_ATTR)
			{
				do
				{	if(xch == '\"' || xch == '\'') { quoteCHAR = xch; break; }
					else if(xch == '=') equalCount++;
					else if(xch == ' ' || xch == '	' || xch == '\r') ;
					else if(xch == '\n') { lineCount++; charPos = i; }
					else throwError(1); //Error
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);

				if(equalCount != 1) throwError(1); //Error: Invalid Char
				xdataLen = 0; iFlag = false;
				do
				{	i++;    if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) break;
				    if(xch == quoteCHAR) { curBLK = handleAttr(); iFlag = true; }
					else if(xch == '<') throwError(1); //Error: Invalid Char
					else { xdata[xdataLen] = xch;	xdataLen++; }
				} while(true);
			}


			else if(curBLK == BLK_TAGBEGIN) {
				if(xch != '<') {
					if(docBegin) throwError(1); //Error: Invalid char
					else docBegin = true;
				}
				else {
					i++;
					iFlag = true;
					curBLK = BLK_TAGNAME;

					do
					{	if(i > charsRead) refreshxBuffer();
						xch = xbuf[i];
						if(iFlag) {
							if(xch == '/' || xch == '?') tagQualifier=xch;
							else if(xch == '!') curBLK = BLK_MISC;
							else break;
							i++; iFlag = false;
						}
						else break;
					} while(true);

					xdataLen=0;
					continue CoreLoop;
				}
			}


			else if(curBLK == BLK_TAGEND) {
				if(xch == '>') iFlag = false;
				else if(xch == '/' || xch == '?') { tagQualifier = xch; iFlag = true; }
				else throwError(1); //Error: Invalid char
				i++;

				do
				{	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) {
						if(xch == '>') { i++; iFlag = false; closeTAG(); }
						else throwError(1);	//Error: Invalid Char. Expected '>'
					}
					else break;
				} while(true);

				if(tagOpen) {
					tagOpen = false; xdataLen = 0;
//System.out.print(">");
				}
				else throwError(2); //Error: null tag name
				curBLK = BLK_TEXTDATA;
				continue CoreLoop;
			}


			else if(curBLK == BLK_TEXTDATA)
			{
				do
				{	if(xch == '<')  { curBLK = handleChrData(); continue CoreLoop; }
					else { xdata[xdataLen] = xch; xdataLen++; }
if(xch == '\n') { lineCount++; charPos = i; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);
			}


			else if(curBLK == BLK_CHRDATA)
			{
//if(markCHAR == '-') System.out.print("\n<!--");
//else System.out.print("\n<![CDATA[");
				iFlag = false; markCount = 0;
				do
				{	if(markCount > 1 && xch == '>') iFlag = true;
					if(xch == markCHAR) markCount++;
					else markCount = 0;
					xdata[xdataLen] = xch; xdataLen++;
if(xch == '\n') { lineCount++; charPos = i; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) { curBLK = handleChrData(); break; }
				} while(true);
				curBLK = BLK_TEXTDATA;
			}


			else if(curBLK == BLK_MISC)
			{
				curBLK = BLK_CHRDATA;
				if(xch == '-') { cMarkup = commentMarkup; markCHAR = '-'; }
				else if(xch == 'E') { cMarkup = entityMarkup; curBLK = BLK_ENTITY; }
				else if(xch == '[') { cMarkup = cdataMarkup; markCHAR = ']'; }
				else if(xch == 'D') { cMarkup = doctypeMarkup; curBLK = BLK_DOCTYPE1; }
				else throwError(1); //Error: Invalid Char

				markupPos = 0;  markCount = cMarkup.length-1;
				do
				{	i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(markupPos > markCount) break;
					if(xch == cMarkup[markupPos]) markupPos++;
					else throwError(1);	//Error: Invaid Tag Char
				} while(true);
				continue CoreLoop;
			}


			else if(curBLK == BLK_ENTITY)
			{
				do
				{	if(xch == '\"' || xch == '\'') { markCHAR = xch; break; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);

				iFlag = false; markCount = 0;
				do
				{	i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) {
						if(markCount > 0) break;
						if(xch == '>') markCount++;
						else if(xch == ' ' || xch == '	' || xch == '\r') ;
						else if(xch == '\n') { lineCount++; charPos = i; }
						else throwError(1);	//Error: Invalid Char
					}
					else if(xch == markCHAR)
					iFlag = true;
				} while(true);
				curBLK = BLK_DOCTYPE2;
			}


			else if(curBLK == BLK_DOCTYPE2)
			{
				iFlag = false; markCount = 0;
				do
				{	if(iFlag) {
						if(markCount > 0) break;
						if(xch == markCHAR) markCount++;
						else if(markCHAR == '>' && (xch == ' ' || xch == '	' || xch == '\r' || xch == '\n'))
							{ if(xch == '\n') { lineCount++; charPos = i; } }
						else throwError(1); //Error Invalid Char
					}
					else {
						if(xch == '<') { markCHAR = '!';  curBLK = BLK_MISC; iFlag = true; }
						else if(xch == ' ' || xch == '	' || xch == '\r') ;
						else if(xch == '\n') { lineCount++; charPos = i; }
						else if(xch == ']') { markCHAR = '>';  curBLK = BLK_TAGBEGIN; iFlag = true; }
						else throwError(1); //Error: Invalid Char
					}
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);
			}


			else if(curBLK == BLK_DOCTYPE1)
			{
				iFlag = false; markCount = 0;
				do
				{	if(xch == '[') { curBLK = BLK_DOCTYPE2; iFlag = true; }
					else if(xch == '>') { curBLK = BLK_TAGBEGIN; iFlag = true; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) break;
				} while(true);

			}


			else if(curBLK == BLK_PI)
			{
				iFlag = false; markCount = 0;
				do
				{	if(markCount > 0 && xch == '>') iFlag = true;
					else if(xch == '?') markCount++;
					else markCount = 0;
					xdata[xdataLen] = xch; xdataLen++;
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) { curBLK = handleChrData(); tagOpen = false; break; }
				} while(true);
			}


			//generic loop for parsing whitespaces
			do
			{	if(xch == ' ' || xch == '	' || xch == '\r') ;
				else if(xch == '\n') { lineCount++; charPos = i; }
				else break;
				i++;
				if(i > charsRead) refreshxBuffer();
				xch = xbuf[i];
			} while(true);

		} while(true); //End of CoreLoop

		} while(true); //End of ParserLoop


		//--------------------------------------------------------------
		//                    end of parser logic
		//--------------------------------------------------------------
		} catch(IOException e) { System.out.println(e.toString()); return null; }
		  catch(ArrayIndexOutOfBoundsException ei) { }
		  catch(InstantiationException ee) { return null; }

		return null;
	}


	void refreshxBuffer() throws IOException
	{
		charPos -= i-1;
		i = 0;

		do {
			charsRead = xstream.read(xbuf, 0, xBufSize) -1;
			if(charsRead == 0) continue;
			else if(charsRead < -1) xbuf = new char[0];
			return;
		} while(true);
	}


	short handleTAG() throws InstantiationException
	{
		String nStr = new String(xdata, 0, xdataLen);
		short newBlk = BLK_TAGNAME;

		if(tagOpen) {
			newBlk = BLK_ATTR;
		}
		else {
			if(tagQualifier == '/') {
				if(cNodeOpen && nStr.equals(cNodeName)) cNodeOpen = false;
				else throwError(3);
				newBlk = BLK_TAGEND;
			}
			else if(tagQualifier == '?') {
				newBlk = BLK_PI;
			}
			else {
				XMLNode xn  = new XMLNode();
				xn.nodeType = Node.ELEMENT_NODE;
				xn.ownerDoc = Doc;
				xn.nodeData = new Object[Xdef.ELEMENT_DATA_COUNT];
				xn.nodeData[Xdef.IELEMENT_TAGNAME] = nStr;
			}
			tagOpen = true;
		}

		xdataLen = 0;
		tagQualifier = ' ';
		return newBlk;
	}

	void closeTAG() throws InstantiationException
	{
		if(tagOpen) ;	//Close current Node
		else throwError(1);	//Error: Invalid Char at i-1
		tagQualifier = ' ';
	}

	short handleAttr()
	{
		xdataLen = 0;
		equalCount = 0;
		return BLK_TAGNAME;
	}

	short handleChrData()
	{
//		if(curBLK == BLK_PI) System.out.print(" ");
//		System.out.print(String.copyValueOf(xdata, 0, xdataLen));
		xdataLen = 0;
		return BLK_TAGBEGIN;
	}

	void throwError(int errNo) throws InstantiationException
	{	String errPos = " line "+lineCount+" char "+(i-charPos);

		switch(errNo)
		{
			case 1:
				System.out.println("Malformed Document Error: Invalid Character at"+errPos);
				break;

			case 2:
				System.out.println("Null Tag Name at"+errPos);
				break;

			default:
				System.out.println("Parse Error at"+errPos);
		}

		throw new InstantiationException();
	}



	public static void main(String[] args)
	{
		Parser tp = new Parser();
		if(args.length < 1 || args[0].equals("1"))
			tp.load("c:\\Exml\\Appntmts.xml");
		else if(args[0].equals("2"))
			tp.load("c:\\Exml\\REC-xml-19980210.xml");
		else
			tp.load("c:\\Exml\\Large.xml");
		System.exit(0);
	}
}
