import org.w3c.dom.*;
import java.io.*;


public class Parser
{
	XMLNode doc;
	InputStreamReader xstream;

	static final short BLK_WHITESPACE=0;
	static final short BLK_ELEMENTBEGIN=1;
	static final short BLK_ELEMENTEND=2;
	static final short BLK_TAGNAMEBEGIN=3;
	static final short BLK_TAGNAME=4;
	static final short BLK_PI=5;
	static final short BLK_CHRDATA=6;
	static final short BLK_TEXTDATA=7;
	static final short BLK_CDATATAG=8;
	static final short BLK_CDATA=9;


	static final short CTX_ELEMENT=0;
	static final short CTX_ATTRNAME=1;
	static final short CTX_ATTRVALUE=2;
	static final short CTX_TEXT=3;
	static final short CTX_CDATA=4;
	static final short CTX_PI=5;
	static final short CTX_EDATA=6;
	static final short CTX_DOCTYPE=7;
	static final short CTX_ENTITY=8;
	static final short CTX_NOTATION=9;
	static final short CTX_COMMENT=10;

	short curCTX = CTX_ELEMENT;
	short curBLK = BLK_WHITESPACE;
	short blkFallThrough = BLK_ELEMENTBEGIN;

	int xBufSize=4096, dataBufSize=4096;
	int xdataLen=0;
	char[] xbuf;
	char[] xdata;

	char[] cdtagData = { 'C', 'D', 'A', 'T', 'A', '[' };
	short  cdtagPos  = 0;

	public Document load(String xmlFile)
	{
		return null;
	}

	Document parse(InputStream xs)
	{
		int i, charsRead;
		int cMark1 = -9, cMark2 = -9;
		char xch, markCHAR=0;

		xstream = new InputStreamReader(xs);
		xbuf  = new char[xBufSize];
		xdata = new char[dataBufSize];

		doc = new XMLNode();
		doc.nodeType = Node.DOCUMENT_NODE;
		doc.nodeData = new Object[Xdef.DOCUMENT_DATA_COUNT];

		try
		{

		OuterLoop:
		do
		{
		charsRead = refreshxBuffer();
		if(charsRead < -1) break;	//End of Stream
		i=0;  xch = xbuf[0];

		DataLoop:
		do
		{
		//--------------------------------------------------------------
		//             DataLoop Logic - Also the Parser Logic
		//--------------------------------------------------------------

		//generic loop for parsing whitespaces
		while(xch == ' ' || xch == '	' || xch == '\r' || xch == '\n')
		{	i++;
			if(i > charsRead) {
				charsRead = refreshxBuffer();
				if(charsRead < -1) break OuterLoop;	//End of Stream
				i = 0;
			}
			xch = xbuf[i];
		}

		//inner loops for parsing Tags, Attributes, Text & CDATA
		CoreDataRefreshLoop:
		do
		{
		CoreDataLoop:
		do
		{
			//TagName begin : most frequent case
			if(curBLK == BLK_TAGNAMEBEGIN) {
				if((xch > '@' && xch < '[') || (xch > '`' && xch < '{') || xch == ':' || xch == '_')
				{
					xdata[0] = xch;
					xdataLen = 1;
					curBLK = BLK_TAGNAME;
				}
				else if(curCTX == CTX_ELEMENT)
				{
					if(xch == '!')
					{
						if(curCTX == CTX_EDATA) ;//Error: double '!' char
						curCTX = CTX_EDATA;
					}
					else if(xch == '?')
					{
						if(curCTX == CTX_PI) ;	//Error: double '?' char
						curCTX = CTX_PI;
					}
					else if(xch == '[')
					{
						if(curCTX != CTX_EDATA) ;	//Error: Invalid char
						curCTX = CTX_CDATA;
						curBLK = BLK_CDATATAG;
						cdtagPos = 0;
					}
					else if(xch == '-')
					{
						if(curCTX != CTX_EDATA) ;	//Error: Invalid char
						curCTX = CTX_COMMENT;
						curBLK = BLK_CDATATAG;
					}
					else ; //Error: Invalid Tag Name Begin Char
				}
				else ; //Error: Invalid Tag Name Begin Char

				i++;
				if(i > charsRead) break CoreDataLoop;
				else xch = xbuf[i];
			}


			else if(curBLK == BLK_TAGNAME) {
				while((xch > '@' && xch < '[') || (xch > '`' && xch < '{') ||
					(xch > '/' && xch <= '9') || xch == ':' || xch == '_' || xch == '-')
				{
					xdata[xdataLen] = xch;
					xdataLen++;
					i++;
					if(i > charsRead) break CoreDataLoop;
					else xch = xbuf[i];
				}

				curBLK = BLK_TAGNAMEBEGIN;
				continue DataLoop;

				if(xch == '>' && curCTX == CTX_ELEMENT)
				{
					openTAG();
					xdataLen = 0;
					curBLK = BLK_TEXTDATA;
				}
				else ; //Error: Invalid Tag Name Char
				i++;
				continue;
			}


			else if(curBLK == BLK_ELEMENTBEGIN) {
				if(xch == '<') {
					curBLK = BLK_TAGNAMEBEGIN;
					xdataLen=0;
				}
				else ;	//Error: Invaid Char expected '<'
				i++;
				continue;
			}


			else if(curBLK == BLK_CHRDATA)
			{
				if(xch == markCHAR)
				{
					if(cMark2 > -1) { cMark1 = cMark2; cMark2 = xdataLen; }
					else if(cMark1 > -1) cMark2 = xdataLen;
					else cMark1 = xdataLen;
				}
				else if(xch == '>')
				{
					if(xdataLen == cMark2+1 && cMark2 == cMark1+1) {
						//End Entity or Comment Node
						curBLK = BLK_TEXTDATA;
						xdataLen = 0;
						continue;
					}
				}
				xdata[xdataLen] = xch;
				xdataLen++; i++;
			}


			else if(curBLK == BLK_TEXTDATA)
			{
				if(xch == ' ' || xch == '	' || xch == '\r' || xch == '\n')
					{ i++; continue; }
				else if(xch == '<')
				{
					//Begin new Element
				}
				xdata[xdataLen] = xch;
				xdataLen++; i++;
			}


			else if(curBLK == BLK_CDATATAG) {
				if(curCTX == CTX_COMMENT) {
					if(xch == '-') {
						curBLK = BLK_CHRDATA;
						markCHAR = '-';
						cMark1 = -9; cMark2 = -9;
					}
					else ;//Error: Invaid Tag Char
				}
				else {
					if(xch == cdtagData[cdtagPos]) { i++; cdtagPos++; }
					else ;	//Error: Invaid Tag Char
					if(cdtagPos > 5) {
						curBLK = BLK_CHRDATA;
						markCHAR = ']';
						cMark1 = -9; cMark2 = -9;
					}
				}
				continue;
			}

		} while(true);

		charsRead = refreshxBuffer();
		if(charsRead < -1) break OuterLoop;	//End of Stream
		i=0;  xch = xbuf[0];

		} while(true);


		//--------------------------------------------------------------
		//                    end of parser logic
		//--------------------------------------------------------------
		} while(true);
		} while(true);
		} catch(IOException e) { System.out.println(e.toString()); return null; }

		return null;
	}


	int refreshxBuffer() throws IOException
	{
		while(true)
		if(xstream.ready())
			return xstream.read(xbuf, 0, xBufSize) -1;
		else continue;
	}

	void openTAG()
	{

	}





	public static void main(String[] args)
	{
	int i=0, j=0;

	MainLoop:
	while(true) {

	Loop1:
	while(true) {
		if(i > 8) break Loop1;
		System.out.println("In Loop 1");
		i++;
		//continue Loop1;
		break;
	}

	Loop2:
	while(true) {
		if(j > 4) break MainLoop;
		System.out.println("In Loop 2");
		j++;
		continue MainLoop;
	}

	}



	}
}
