//import org.w3c.dom.*;

public class XMLNode
implements Document, DocumentType, DocumentFragment, DOMImplementation,
           Node, Attr, CDATASection, CharacterData, Comment, Element,
           Entity, EntityReference, NamedNodeMap, NodeList, Notation,
           ProcessingInstruction, Text
{
	int     nodeDepth = -99;
	short   nodeType = Node.ELEMENT_NODE;
	String  nodeChr;
	XMLNode ownerDoc;
	XMLNode parentNode, nextNode, prevNode, nextSibling, prevSibling;
	Object[] nodeData;  //xn.nodeData[Xdef.IELEMENT_TAGNAME] = nStr;

	//static int seq=-1;
	//int nodeID=seq++;

	//Document Interface
	public DocumentType getDoctype() { return null; }

	public DOMImplementation getImplementation() { return this; }

	public Element getDocumentElement()
	{
		return (Element)nodeData[Xdef.IDOCUMENT_ROOT];
	}

	public Element createElement(String tagName) throws DOMException
	{
		XMLNode tn  = new XMLNode();
		tn.ownerDoc = this;
		tn.nodeChr  = tagName;
		return (Element)tn;
	}

	public DocumentFragment createDocumentFragment()
	{
		XMLNode tn  = new XMLNode();
		tn.nodeType = Node.DOCUMENT_FRAGMENT_NODE;
		tn.ownerDoc = this;
		return (DocumentFragment)tn;
	}

	public Text createTextNode(String data)
	{
		XMLNode tn  = new XMLNode();
		tn.nodeType = Node.TEXT_NODE;
		tn.ownerDoc = this;
		tn.nodeChr  = data;
		return (Text)tn;
	}

	public Comment createComment(String data)
	{
		XMLNode tn  = new XMLNode();
		tn.nodeType = Node.COMMENT_NODE;
		tn.ownerDoc = this;
		tn.nodeChr  = data;
		return (Comment)tn;
	}

	public CDATASection createCDATASection(String data)
													 throws DOMException { return null; }
	public ProcessingInstruction createProcessingInstruction(String target,
																			 String data)
																			 throws DOMException { return null; }

	public Attr createAttribute(String name) throws DOMException
	{
		XMLNode tn  = new XMLNode();
		tn.nodeType = Node.ATTRIBUTE_NODE;
		tn.ownerDoc = this;
		tn.nodeChr  = name;
		return (Attr)tn;
	}

	public EntityReference createEntityReference(String name)
															 throws DOMException { return null; }

	//Common implmentation for Document & Element Interface
	public NodeList getElementsByTagName(String tagname)
	{
		XMLNode tn  = new XMLNode();
		tn.nodeType = Xdef.NODELIST_ELEMENTS;
		tn.nodeChr  = tagname;
		if(nodeType == Node.ELEMENT_NODE) tn.parentNode = this;
		else tn.parentNode = (XMLNode)(nodeData[Xdef.IDOCUMENT_ROOT]);
		tn.ownerDoc  = tn.parentNode.ownerDoc;
		tn.nextNode  = tn.parentNode;
		tn.nodeDepth = tn.parentNode.nodeDepth;
		tn.nodeData = new Object[1];
		tn.nodeData[0] = new int[] { -1, -1, -1 };	//0-ModifiedCount, 1-Length, 2-curIndex
		return tn;
	}


	//DocumentType Interface
	//public String getName();  --- Part of Attr Interface
	public NamedNodeMap getEntities() { return null; }
	public NamedNodeMap getNotations() { return null; }

	//DOMImplementation Interface
	public boolean hasFeature(String feature, String version) { return false; }


	//Node Interface
	public String getNodeName()
	{
		if(nodeType == Node.ELEMENT_NODE) return nodeChr;
		else if(nodeType == Node.ATTRIBUTE_NODE) return nodeChr;
		else if(nodeType == Node.TEXT_NODE) return "Text";

		return null;
	}

	public String getNodeValue() throws DOMException
	{
		if(nodeType == Node.TEXT_NODE ||
			nodeType == Node.CDATA_SECTION_NODE ||
			nodeType == Node.COMMENT_NODE)
			  return nodeChr;
		else if(nodeType == Node.ATTRIBUTE_NODE && nodeData !=null) return (String)(nodeData[0]);
		else return null;
	}

	public void setNodeValue(String nodeValue) throws DOMException
	{
		if(nodeType == Node.TEXT_NODE ||
			nodeType == Node.CDATA_SECTION_NODE ||
			nodeType == Node.COMMENT_NODE)
				nodeChr = nodeValue;
		else if(nodeType == Node.ATTRIBUTE_NODE)
		{
			if(nodeData == null) nodeData = new Object[1];
			nodeData[0] = nodeValue;
		}
	}

	public short getNodeType() { return nodeType; }

	public Node getParentNode() { return parentNode; }

	public NodeList getChildNodes()
	{
		if(nodeType == Node.ELEMENT_NODE || nodeType == Node.DOCUMENT_FRAGMENT_NODE)
		{
		XMLNode tn  = new XMLNode();
		tn.nodeType = Xdef.NODELIST_CHILDREN;
		tn.ownerDoc = this.ownerDoc;
		tn.parentNode = this;
		tn.nextNode  = (XMLNode)this.getFirstChild();
		tn.nodeDepth = this.nodeDepth;
		tn.nodeData = new Object[1];
		tn.nodeData[0] = new int[] { -1, -1, 0 };	//0-ModifiedCount, 1-Length, 2-curIndex
		return tn;
		}
		else return null;
	}

	public Node getFirstChild()
	{
		if(nextNode != null && nextNode.nodeDepth > nodeDepth) return nextNode;
		else return null;
	}

	public Node getLastChild()
	{
		XMLNode tNode = nextNode;
		if(tNode == null || tNode.nodeDepth <= nodeDepth) return null;	//No children
		while(tNode.nextSibling != null) tNode=tNode.nextSibling;
		return tNode;
	}

	public Node getPreviousSibling() { return prevSibling; }

	public Node getNextSibling() { return nextSibling; }

	public NamedNodeMap getAttributes()
	{
		if(nodeType != Node.ELEMENT_NODE) return null;

		XMLNode tn  = new XMLNode();
		tn.nodeType = Xdef.NAMEDNODEMAP_ATTR;
		tn.ownerDoc = this.ownerDoc;
		tn.parentNode = this;
		tn.nextNode   = this;
		tn.nodeDepth  = this.nodeDepth;
		tn.nodeData = new Object[1];
		tn.nodeData[0] = new int[] { -1, -1, -1 };	//0-ModifiedCount, 1-Length, 2-curIndex
		return tn;
	}

	public Document getOwnerDocument() { return ownerDoc; }

	public Node insertBefore(Node newChild, Node refChild) throws DOMException
	{
		//To be implemented later.......
		return null;
	}

	public Node replaceChild(Node newChild, Node oldChild) throws DOMException
	{
		//To be implemented later.......
		return null;
	}

	public Node removeChild(Node oldChild) throws DOMException
	{
		XMLNode eNode, rNode=this.nextNode;

		//make sure rNode is first child
		if(rNode == null || rNode.nodeDepth <= nodeDepth) return null; //No Children

		for( ; rNode != null; rNode=rNode.nextSibling)
		 if(rNode == oldChild)
		 {	//pop the node out
		 	eNode = getLastNode((XMLNode)oldChild);
			rNode.prevNode.nextNode = eNode.nextNode;
			if(eNode.nextNode != null) eNode.nextNode.prevNode = rNode.prevNode;
			if(rNode.prevSibling != null) rNode.prevSibling.nextSibling = rNode.nextSibling;
			if(rNode.nextSibling != null) rNode.nextSibling.prevSibling = rNode.prevSibling;
			return rNode;
		 }
		return null;
	}

	//First version: doesn't check nodeType for legal parent-child relation
	public Node appendChild(Node newChild) throws DOMException
	{
		XMLNode tNode, nNode=(XMLNode)newChild, neNode, teNode;

		//if not a fresh node, perform DOM checks
		if(nNode.nodeDepth > 0)
		{	if(nNode.ownerDoc != ownerDoc) throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, "Illegal owner Doc");
			if(isAncestor(nNode)) throw new DOMException(DOMException.HIERARCHY_REQUEST_ERR, "Node is Ancestor");
			removeChild(newChild); //remove if the node already a child
		}
		//prepare for adding
		updateDepth(nNode, (nodeDepth < 0) ? 1 : nodeDepth+1);
		nNode.parentNode = this;

		tNode  = (XMLNode)getLastChild();//get last child of this node
		teNode = getLastNode(this);		//get last node in the last child's tree
		neNode = getLastNode(nNode);		//get last node in the new child's tree
		//Do all the book-keeping of XMLNodes
		neNode.nextNode = teNode.nextNode;
		if(teNode.nextNode != null) teNode.nextNode.prevNode = neNode;
		teNode.nextNode = nNode;
		nNode.prevNode  = teNode;
		if(tNode != null) tNode.nextSibling = nNode;
		nNode.prevSibling = tNode;
		return newChild;
	}

	public boolean hasChildNodes()
	{
		if(nextNode != null && nextNode.nodeDepth > nodeDepth) return true;
		else return false;
	}

	public Node cloneNode(boolean deep) throws DOMException
	{
		//To be implemented later.......
		return null;
	}

	//Node Utils
	boolean isAncestor(XMLNode tN)
	{
		XMLNode tNode=this;
		while(tNode != null) { if(tNode == tN) return true; tNode = tNode.parentNode; }
		return false;
	}
	XMLNode getLastNode(XMLNode tN)
	{
		XMLNode tNode = tN;
		while(tNode.nextNode != null && tNode.nextNode.nodeDepth > tN.nodeDepth) tNode=tNode.nextNode;
		return tNode;
	}
	void updateDepth(XMLNode tN, int D)
	{
		XMLNode tNode = tN.nextNode;
		D = (tN.nodeDepth < 0) ? D : D-tN.nodeDepth;
		while(tNode != null && tNode.nodeDepth > tN.nodeDepth)
			{ tNode.nodeDepth += D; tNode=tNode.nextNode; }
		tN.nodeDepth = D;
	}

	//Element Interface
	public String getTagName() { return nodeChr; }

	public String getAttribute(String name)
	{
		if(nodeData == null) return null;
		int attrCount = ((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = ((String[])nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = ((String[])nodeData[Xdef.IELEMENT_ATTRVALUE]);
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(name)) return attrValue[i];
		return null;
	}

	public void setAttribute(String name, String value) throws DOMException
	{
		if(nodeData == null)
		{	Object[] tData = new Object[Xdef.ELEMENT_DATA_COUNT];
			tData[Xdef.IELEMENT_ATTRCOUNT]  = new int[] { 0 };
			tData[Xdef.IELEMENT_ATTRNAME]   = new String[Xdef.ARRAY_RESIZE];
			tData[Xdef.IELEMENT_ATTRVALUE]  = new String[Xdef.ARRAY_RESIZE];
			nodeData = tData;
		}
		int attrCount = ((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = (String[])(nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = (String[])(nodeData[Xdef.IELEMENT_ATTRVALUE]);
		int pos = attrCount;
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(name)) { pos=i; break; }
		if(pos < attrCount) { attrValue[pos] = value; return; }
		if(attrName.length == attrCount)
		{
			String[] attrName2  = new String[attrName.length+Xdef.ARRAY_RESIZE];
			String[] attrValue2 = new String[attrName.length+Xdef.ARRAY_RESIZE];
			System.arraycopy(attrName, 0, attrName2, 0, attrName.length);
			System.arraycopy(attrValue, 0, attrValue2, 0, attrName.length);
			attrName = attrName2;  attrValue = attrValue2;
			nodeData[Xdef.IELEMENT_ATTRNAME]  = attrName;
			nodeData[Xdef.IELEMENT_ATTRVALUE] = attrValue;
		}
		attrName[pos]  = name;
		attrValue[pos] = value;
		attrCount++;
		((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0] = attrCount;
	}

	public void removeAttribute(String name) throws DOMException
	{
		if(nodeData == null) return;
		int attrCount = ((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = (String[])(nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = (String[])(nodeData[Xdef.IELEMENT_ATTRVALUE]);
		int pos = -1;
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(name)) { pos=i; break; }
		if(pos < 0) return;
		String[] attrName2  = new String[attrName.length];
		String[] attrValue2 = new String[attrName.length];
		System.arraycopy(attrName, 0, attrName2, 0, pos);
		System.arraycopy(attrName, pos+1, attrName2, pos, attrCount-pos-1);
		System.arraycopy(attrValue, 0, attrValue2, 0, pos);
		System.arraycopy(attrValue, pos+1, attrValue2, 0, attrCount-pos-1);
		nodeData[Xdef.IELEMENT_ATTRNAME]  = attrName2;
		nodeData[Xdef.IELEMENT_ATTRVALUE] = attrValue2;
		attrCount--;
		((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0] = attrCount;
	}

	public Attr getAttributeNode(String name)
	{
		if(nodeData == null) return null;
		int attrCount = ((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = ((String[])nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = ((String[])nodeData[Xdef.IELEMENT_ATTRVALUE]);
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(name))
			{	XMLNode tn  = new XMLNode();
				tn.nodeType = Node.ATTRIBUTE_NODE;
				tn.ownerDoc = this.ownerDoc;
				tn.parentNode = this;
				tn.nodeChr  = name;
				tn.nodeData = new Object[1];
				tn.nodeData[0] = new String(attrValue[i]);
				return tn;
			}
		return null;
	}

	public Attr setAttributeNode(Attr newAttr) throws DOMException
	{
		if(nodeData == null)
		{	Object[] tData = new Object[Xdef.ELEMENT_DATA_COUNT];
			tData[Xdef.IELEMENT_ATTRCOUNT]  = new int[] { 0 };
			tData[Xdef.IELEMENT_ATTRNAME]   = new String[Xdef.ARRAY_RESIZE];
			tData[Xdef.IELEMENT_ATTRVALUE]  = new String[Xdef.ARRAY_RESIZE];
			nodeData = tData;
		}
		int attrCount = ((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = (String[])(nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = (String[])(nodeData[Xdef.IELEMENT_ATTRVALUE]);
		int pos = attrCount;
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(((XMLNode)newAttr).nodeChr)) { pos=i; break; }
		if(pos < attrCount) { attrValue[pos] = (String)(((XMLNode)newAttr).nodeData[0]); return newAttr; }
		if(attrName.length == attrCount)
		{
			String[] attrName2  = new String[attrName.length+Xdef.ARRAY_RESIZE];
			String[] attrValue2 = new String[attrName.length+Xdef.ARRAY_RESIZE];
			System.arraycopy(attrName, 0, attrName2, 0, attrName.length);
			System.arraycopy(attrValue, 0, attrValue2, 0, attrName.length);
			attrName = attrName2;  attrValue = attrValue2;
			nodeData[Xdef.IELEMENT_ATTRNAME]  = attrName;
			nodeData[Xdef.IELEMENT_ATTRVALUE] = attrValue;
		}
		attrName[pos]  = ((XMLNode)newAttr).nodeChr;
		attrValue[pos] = (String)(((XMLNode)newAttr).nodeData[0]);
		attrCount++;
		((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0] = attrCount;
		return newAttr;
	}

	public Attr removeAttributeNode(Attr oldAttr) throws DOMException
	{
		if(nodeData == null) return null;
		int attrCount = ((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = (String[])(nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = (String[])(nodeData[Xdef.IELEMENT_ATTRVALUE]);
		int pos = -1;
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(((XMLNode)oldAttr).nodeChr)) { pos=i; break; }
		if(pos < 0) return null;
		String[] attrName2  = new String[attrName.length];
		String[] attrValue2 = new String[attrName.length];
		System.arraycopy(attrName, 0, attrName2, 0, pos);
		System.arraycopy(attrName, pos+1, attrName2, pos, attrCount-pos-1);
		System.arraycopy(attrValue, 0, attrValue2, 0, pos);
		System.arraycopy(attrValue, pos+1, attrValue2, 0, attrCount-pos-1);
		nodeData[Xdef.IELEMENT_ATTRNAME]  = attrName2;
		nodeData[Xdef.IELEMENT_ATTRVALUE] = attrValue2;
		attrCount--;
		((int[])nodeData[Xdef.IELEMENT_ATTRCOUNT])[0] = attrCount;
		return oldAttr;
	}

	//public NodeList getElementsByTagName(String name) { return null; } //Part of Document Interface

	public void normalize() { }	//To be implemented later


	//Attr Interface
	public String getName() { return nodeChr; }

	public boolean getSpecified()
	{
		if(parentNode == null) return true;
		if(nodeData == null) return false;
		else return true;
	}

	public String getValue()
	{
		if(nodeData == null) return null;
		else return (String)(nodeData[0]);
	}

	public void setValue(String value) throws DOMException
	{
		if(nodeData == null) nodeData = new Object[1];
		nodeData[0] = value;
	}

	//Text Interface
	public Text splitText(int offset) throws DOMException
	{
		return this;
	}

	//NamedNodeMap and NodeList Interface
	public int getLength()	//Common for both NodeList & NamedNodeMap
	{
		if(((int[])nodeData[0])[0] == ((int[])ownerDoc.nodeData[Xdef.IDOCUMENT_MODIFIEDCOUNT])[0])
			return ((int[])nodeData[0])[1];

		int Len=0;
		XMLNode tNode;

		if(nodeType == Xdef.NODELIST_ELEMENTS)
		{	tNode = parentNode.nextNode;
			while(tNode != null && tNode.nodeDepth > nodeDepth)
			{
				if(tNode.nodeType == Node.ELEMENT_NODE &&
					(tNode.nodeChr.equals(nodeChr) || nodeChr.equals("*"))) Len++;
				tNode = tNode.nextNode;
			}
		}
		else if(nodeType == Xdef.NODELIST_CHILDREN)
		{	tNode = nextNode;
			while(tNode != null) {
				Len++; tNode = tNode.nextSibling; }
		}
		else if(nodeType == Xdef.NAMEDNODEMAP_ATTR)
		{
			if(parentNode.nodeData == null) Len=0;
			else Len = ((int[])parentNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		}

		((int[])nodeData[0])[0] = ((int[])ownerDoc.nodeData[Xdef.IDOCUMENT_MODIFIEDCOUNT])[0];
		((int[])nodeData[0])[1] = Len;
		return Len;
	}

	public Node getNamedItem(String name)
	{
		if(parentNode.nodeData == null) return null;
		int attrCount = ((int[])parentNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = ((String[])parentNode.nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = ((String[])parentNode.nodeData[Xdef.IELEMENT_ATTRVALUE]);
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(name))
			{	XMLNode tn  = new XMLNode();
				tn.nodeType = Node.ATTRIBUTE_NODE;
				tn.ownerDoc = this.ownerDoc;
				tn.parentNode = parentNode;
				tn.nodeChr  = name;
				tn.nodeData = new Object[1];
				tn.nodeData[0] = new String(attrValue[i]);
				return tn;
			}
		return null;
	}

	public Node setNamedItem(Node arg) throws DOMException
	{
		if(parentNode.nodeData == null)
		{	Object[] tData = new Object[Xdef.ELEMENT_DATA_COUNT];
			tData[Xdef.IELEMENT_ATTRCOUNT]  = new int[] { 0 };
			tData[Xdef.IELEMENT_ATTRNAME]   = new String[Xdef.ARRAY_RESIZE];
			tData[Xdef.IELEMENT_ATTRVALUE]  = new String[Xdef.ARRAY_RESIZE];
			parentNode.nodeData = tData;
		}
		int attrCount = ((int[])parentNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = (String[])(parentNode.nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = (String[])(parentNode.nodeData[Xdef.IELEMENT_ATTRVALUE]);
		int pos = attrCount;
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(((XMLNode)arg).nodeChr)) { pos=i; break; }
		if(pos < attrCount) { attrValue[pos] = (String)(((XMLNode)arg).nodeData[0]); return arg; }
		if(attrName.length == attrCount)
		{
			String[] attrName2  = new String[attrName.length+Xdef.ARRAY_RESIZE];
			String[] attrValue2 = new String[attrName.length+Xdef.ARRAY_RESIZE];
			System.arraycopy(attrName, 0, attrName2, 0, attrName.length);
			System.arraycopy(attrValue, 0, attrValue2, 0, attrName.length);
			attrName = attrName2;  attrValue = attrValue2;
			parentNode.nodeData[Xdef.IELEMENT_ATTRNAME]  = attrName;
			parentNode.nodeData[Xdef.IELEMENT_ATTRVALUE] = attrValue;
		}
		attrName[pos]  = ((XMLNode)arg).nodeChr;
		attrValue[pos] = (String)(((XMLNode)arg).nodeData[0]);
		attrCount++;
		((int[])parentNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0] = attrCount;
		return arg;
	}

	public Node removeNamedItem(String name) throws DOMException
	{
		if(parentNode.nodeData == null) return null;
		int attrCount = ((int[])parentNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
		String[] attrName  = (String[])(parentNode.nodeData[Xdef.IELEMENT_ATTRNAME]);
		String[] attrValue = (String[])(parentNode.nodeData[Xdef.IELEMENT_ATTRVALUE]);
		int pos = -1;
		for(int i=0; i < attrCount; i++)
			if(attrName[i].equals(name)) { pos=i; break; }
		if(pos < 0) return null;
		String[] attrName2  = new String[attrName.length];
		String[] attrValue2 = new String[attrName.length];
		System.arraycopy(attrName, 0, attrName2, 0, pos);
		System.arraycopy(attrName, pos+1, attrName2, pos, attrCount-pos-1);
		System.arraycopy(attrValue, 0, attrValue2, 0, pos);
		System.arraycopy(attrValue, pos+1, attrValue2, 0, attrCount-pos-1);
		parentNode.nodeData[Xdef.IELEMENT_ATTRNAME]  = attrName2;
		parentNode.nodeData[Xdef.IELEMENT_ATTRVALUE] = attrValue2;
		attrCount--;
		((int[])parentNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0] = attrCount;
		return null;	//*** Temp but should actually return a node ***
	}

	public Node item(int index)	//Common for both NodeList & NamedNodeMap
	{
		XMLNode tNode;
		int incr, step=0, delta=index-((int[])nodeData[0])[2];

		if(nodeType == Xdef.NODELIST_ELEMENTS)
		{	if(step < delta) { incr = 1; tNode = nextNode.nextNode; }
			else if(step > delta) { incr = -1; tNode = nextNode.prevNode; }
			else return nextNode;
			while(tNode != null && tNode.nodeDepth > nodeDepth)
			{	if(tNode.nodeType == Node.ELEMENT_NODE &&
					(tNode.nodeChr.equals(nodeChr) || nodeChr.equals("*"))) step += incr;
				if(step == delta) { nextNode=tNode; ((int[])nodeData[0])[2]=index; return tNode; }
				if(incr > 0) tNode = tNode.nextNode;
				else tNode = tNode.prevNode;
			}
		}
		else if(nodeType == Xdef.NODELIST_CHILDREN)
		{	if(delta == 0) return nextNode;
			tNode = nextNode;
			if(step < delta)
				while(tNode != null && step != delta) { tNode=tNode.nextSibling; step++; }
			else	//step > delta
				while(tNode != null && step != delta) { tNode=tNode.prevSibling; step--; }
			if(tNode != null) { nextNode=tNode; ((int[])nodeData[0])[2]=index; return tNode; }
		}
		else if(nodeType == Xdef.NAMEDNODEMAP_ATTR)
		{	if(parentNode.nodeData == null) return null;
			int attrCount = ((int[])parentNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
			if(index >= attrCount || index < 0) return null;
			String[] attrName  = ((String[])parentNode.nodeData[Xdef.IELEMENT_ATTRNAME]);
			String[] attrValue = ((String[])parentNode.nodeData[Xdef.IELEMENT_ATTRVALUE]);
			XMLNode tn  = new XMLNode();
			tn.nodeType = Node.ATTRIBUTE_NODE;
			tn.ownerDoc = this.ownerDoc;
			tn.parentNode = parentNode;
			tn.nodeChr  = attrName[index];
			tn.nodeData = new Object[1];
			tn.nodeData[0] = new String(attrValue[index]);
			return tn;
		}

		return null;
	}


	//CharacterData and PI(partial) Interface
	public String getData()
								 throws DOMException { return null; }  //PI Interface Also
	public void setData(String data)
								 throws DOMException { }  //PI Interface Also
	//public int getLength() { return 0; }	//see NamedNodeMap & NodeList
	public String substringData(int offset,
									  int count)
									  throws DOMException { return null; }
	public void appendData(String arg)
								throws DOMException { }
	public void insertData(int offset,
								String arg)
								throws DOMException { }
	public void deleteData(int offset,
								int count)
								throws DOMException { }
	public void replaceData(int offset,
								 int count,
								 String arg)
								 throws DOMException { }

	//Entity and Notation Interface
	public String getPublicId() { return null; }	//Also for Notation Interface
	public String getSystemId() { return null; }	//Also for Notation Interface
	public String getNotationName() { return null; }

	//ProcessingInstruction Interface
	public String getTarget() { return null; }
}
