//import org.w3c.dom.*;
import java.io.*;

public class XMLParser
{
	XMLNode Doc, cNode=null, sNode=null, eNode=null, tNode;
	String cNodeName = "-99N";
	int cNodeDepth = -99;
	int attrCount=0;
	String[] attrName  = new String[Xdef.MAX_ATTRIBUTES];
	String[] attrValue = new String[Xdef.MAX_ATTRIBUTES];

	InputStreamReader xstream;
	static final int xBufSize=8192, dataBufSize=16384;	//Max size of any node limited to dataBufSize
	int i=0, charsRead, xdataLen=0, lineCount, charPos=0, equalCount=0;
	char[] xbuf;
	char[] xdata;

	short curBLK = Xdef.BLK_TAGBEGIN;
	boolean tagOpen=false;

	char[] cdataMarkup = { 'C', 'D', 'A', 'T', 'A', '[' };
	char[] entityMarkup = { 'N', 'T', 'I', 'T', 'Y' };
	char[] doctypeMarkup = { 'O', 'C', 'T', 'Y', 'P', 'E' };
	char[] commentMarkup = { '-' };
	short markupPos  = 0;
	char  markCHAR=' ', quoteCHAR, tagQualifier=' ';

	private void initParser()
	{
		Doc=null; cNode=null; sNode=null; eNode=null; tNode=null;
		cNodeName = "-99N";
		cNodeDepth = -99;
		attrCount=0;

		xstream=null;
		i=0; xdataLen=0; lineCount=0; charPos=0; equalCount=0;
		curBLK = Xdef.BLK_TAGBEGIN;
		tagOpen=false;
		markupPos = 0;
		markCHAR  =' '; tagQualifier=' ';
	}

	public Document getDocument() { return Doc; }

	public Document createDocument(String doctype)
	{
		XMLNode doc  = new XMLNode();
		doc.nodeType = Node.DOCUMENT_NODE;
		doc.nodeChr  = "#xDocument";
		doc.nodeData = new Object[Xdef.DOCUMENT_DATA_COUNT];
		doc.nodeData[Xdef.IDOCUMENT_MODIFIEDCOUNT] = new int[] { 1 };

		XMLNode root  = new XMLNode();
		root.nodeType = Node.ELEMENT_NODE;
		root.ownerDoc = doc;
		root.nodeChr  = doctype;

		doc.nodeData[Xdef.IDOCUMENT_ROOT] = root;
		return doc;
	}

	public Document parseFile(String xmlFile)
	{
		try {
		long stime = System.currentTimeMillis();
		FileInputStream xf = new FileInputStream(xmlFile);
		//xstream = new InputStreamReader(xf);
		initParser();
		parse(xf);

		/*do  {
		refreshxBuffer();
		System.out.println("CharsRead: "+charsRead);
		//for(int n=0; n < charsRead; n++) markCHAR = xbuf[n];
		} while(charsRead > -1);
		*/
		System.out.println("<expressj parser> Finished parsing "+xmlFile+" in "+(System.currentTimeMillis()-stime)+" ms");
		xf.close();
		} catch(Exception e) { System.out.println(e.toString()); return null; }

		return Doc;
	}

	public Document parseStream(InputStream xmlStream)
	{
		try {
		initParser();
		parse(xmlStream);
		} catch(Exception e) { System.out.println(e.toString()); return null; }
		return Doc;
	}

	public Document parseString(String xml)
	{
			//initParser();
			//parse(xml);
			//return Doc;
			return null;
	}

	public void outputXML(Document doc, OutputStream xs) throws IOException
	{
		byte[] xBytes = getXML(doc).getBytes();
		if(xBytes.length < 1) return;
		int blkSize=8192, blocks=xBytes.length/blkSize, loose=xBytes.length%blkSize;
		for(int i=0; i < blocks; i++) xs.write(xBytes, i*blkSize, blkSize);
		xs.write(xBytes, blocks*blkSize, loose);
	}

	public void saveXML(Document doc, String fileName)  throws IOException
	{
		byte[] xBytes = getXML(doc).getBytes();
		if(xBytes.length < 1) return;
		FileOutputStream xFile = new FileOutputStream(fileName);
		int blkSize=8192, blocks=xBytes.length/blkSize, loose=xBytes.length%blkSize;
		for(int i=0; i < blocks; i++) xFile.write(xBytes, i*blkSize, blkSize);
		xFile.write(xBytes, blocks*blkSize, loose);
		xFile.close();
	}

	public String getXML(Document doc)
	{
		if(doc == null) return "";
		StringBuffer xmlBuf = new StringBuffer();
		XMLNode pNode, nNode, tNode =(XMLNode)(((XMLNode)doc).nodeData[Xdef.IDOCUMENT_ROOT]);
		int i, tDepth=0, nDepth=0;
		boolean hasChildren=false, noSiblings=false;

		do
		{	nNode = tNode.nextNode;
			if(nNode != null)
			{	tDepth = nDepth;
				nDepth = nNode.nodeDepth;
				if(nDepth > tDepth) { hasChildren=true; noSiblings=false; }
				else if(nDepth < tDepth) { hasChildren=false; noSiblings=true; }
				else { hasChildren=false; noSiblings=false; }
			}
			else { tDepth=nDepth; nDepth=0; hasChildren=false; noSiblings=true; }
			xmlBuf.append(getNodeString(tNode, hasChildren));
			if(noSiblings)
				for(i=nDepth, pNode=tNode.parentNode; i < tDepth; i++,pNode=pNode.parentNode)
						xmlBuf.append("</").append(pNode.nodeChr).append(">\n");
			tNode = nNode;
		} while(tNode != null);

		return xmlBuf.toString();
	}

	//public String getXML(Node node)
	//{
	//	return getNodeString((XMLNode)node, true);
	//}

	private String getNodeString(XMLNode tNode, boolean hasChildren)
	{
		StringBuffer xnBuf = new StringBuffer();

		if(tNode.nodeType == Node.ELEMENT_NODE)
		{
			xnBuf.append("<").append(tNode.nodeChr);
			if(tNode.nodeData != null)
			{	int aCount = ((int[])tNode.nodeData[Xdef.IELEMENT_ATTRCOUNT])[0];
				String[] aName  = ((String[])tNode.nodeData[Xdef.IELEMENT_ATTRNAME]);
				String[] aValue = ((String[])tNode.nodeData[Xdef.IELEMENT_ATTRVALUE]);
				for(int i=0; i < aCount; i++)
				xnBuf.append(' ').append(aName[i]).append("=\"").append(aValue[i]).append('\"');
			}
			if(hasChildren) xnBuf.append(">\n");
			else xnBuf.append("/>\n");
		}
		//else if(tNode.nodeType == Node.TEXT_NODE) xnBuf.append(tNode.nodeChr).append('\n');
		else if(tNode.nodeType == Node.TEXT_NODE) xnBuf.append(tNode.nodeChr);
		return xnBuf.toString();
	}

	//Parser Logic implemented here
	void parse(InputStream xs)
	{
		int markCount=0;
		char xch;
		boolean iFlag = false, docBegin=false;
		String curTagName;
		char[] cMarkup=null;

		xstream = new InputStreamReader(xs);
		xbuf  = new char[xBufSize];
		xdata = new char[dataBufSize];

		Doc = new XMLNode();
		Doc.nodeType = Node.DOCUMENT_NODE;
		Doc.nodeChr  = "#xDocument";
		Doc.nodeData = new Object[Xdef.DOCUMENT_DATA_COUNT];
		Doc.nodeData[Xdef.IDOCUMENT_MODIFIEDCOUNT] = new int[] { 1 };

		try
		{

		lineCount = 1;
		refreshxBuffer();
		xch = xbuf[0];

		//--------------------------------------------------------------
		//            ParserLoop - Implements Parser Logic
		//--------------------------------------------------------------

		ParserLoop:
		do
		{

		//Core Loop for parsing Tags, Attributes, Text, CDATA & WhiteSpaces
		CoreLoop:
		do
		{
			//TagName begin : most frequent case
			if(curBLK == Xdef.BLK_TAGNAME) {
				if((xch > '@' && xch < '[') || (xch > '`' && xch < '{') || xch == ':' || xch == '_')
				{
					xdata[0] = xch;
					xdataLen = 1;
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				}
				else { curBLK = Xdef.BLK_TAGEND; continue CoreLoop; }

				while((xch > '@' && xch < '[') || (xch > '`' && xch < '{') ||
					(xch > '/' && xch <= '9') || xch == ':' || xch == '_' || xch == '-')
				{
					xdata[xdataLen] = xch;
					xdataLen++;
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				}

				curBLK = handleTAG();
				//let it fall to whitespace eliminator
			}


			else if(curBLK == Xdef.BLK_ATTR)
			{	//*************************** Bug Alert *******************************
				//The current implementation understands the potential bug
				//for cases when someone specifies attributes like: Name='val1"val2"'.
				//The parser will parse it properly but if you save it, it will be
				//saved as: Name="val1"val2"" !!!which is not well formed. yikes......
				do
				{	if(xch == '\"' || xch == '\'') { quoteCHAR = xch; break; }
					else if(xch == '=') equalCount++;
					else if(xch == ' ' || xch == '	' || xch == '\r') ;
					else if(xch == '\n') { lineCount++; charPos = i; }
					else throwError(1); //Error
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);

				if(equalCount != 1) throwError(1); //Error: Invalid Char
				xdataLen = 0; iFlag = false;
				do
				{	i++;    if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) break;
				    if(xch == quoteCHAR) { curBLK = handleAttr(); iFlag = true; }
					else if(xch == '<') throwError(1); //Error: Invalid Char
					else { xdata[xdataLen] = xch;	xdataLen++; }
				} while(true);
			}


			else if(curBLK == Xdef.BLK_TAGBEGIN) {
				if(xch != '<') {
					if(docBegin) throwError(1); //Error: Invalid char
					else docBegin = true;
				}
				else {
					i++;
					iFlag = true;
					curBLK = Xdef.BLK_TAGNAME;

					do
					{	if(i > charsRead) refreshxBuffer();
						xch = xbuf[i];
						if(iFlag) {
							if(xch == '/' || xch == '?') tagQualifier=xch;
							else if(xch == '!') curBLK = Xdef.BLK_MISC;
							else break;
							i++; iFlag = false;
						}
						else break;
					} while(true);

					xdataLen=0;
					continue CoreLoop;
				}
			}


			else if(curBLK == Xdef.BLK_TAGEND) {
				if(xch == '>') iFlag = false;
				else if(xch == '/' || xch == '?') { tagQualifier = xch; iFlag = true; }
				else throwError(1); //Error: Invalid char
				i++;

				do
				{	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) {
						if(xch == '>') { i++; iFlag = false; closeTAG(); }
						else throwError(1);	//Error: Invalid Char. Expected '>'
					}
					else break;
				} while(true);

				if(tagOpen) {
					tagOpen = false; xdataLen = 0;
				}
				else throwError(2); //Error: null tag name
				curBLK = Xdef.BLK_TEXTDATA;
				//continue CoreLoop;
			}


			else if(curBLK == Xdef.BLK_TEXTDATA)
			{
				do
				{	if(xch == '<')  { curBLK = handleChrData(); continue CoreLoop; }
					else { xdata[xdataLen] = xch; xdataLen++; }
if(xch == '\n') { lineCount++; charPos = i; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);
			}


			else if(curBLK == Xdef.BLK_CHRDATA)
			{
//if(markCHAR == '-') System.out.print("\n<!--");
//else System.out.print("\n<![CDATA[");
				iFlag = false; markCount = 0;
				do
				{	if(markCount > 1 && xch == '>') iFlag = true;
					if(xch == markCHAR) markCount++;
					else markCount = 0;
					xdata[xdataLen] = xch; xdataLen++;
if(xch == '\n') { lineCount++; charPos = i; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) { curBLK = handleChrData(); break; }
				} while(true);
				curBLK = Xdef.BLK_TEXTDATA;
			}


			else if(curBLK == Xdef.BLK_MISC)
			{
				curBLK = Xdef.BLK_CHRDATA;
				if(xch == '-') { cMarkup = commentMarkup; markCHAR = '-'; }
				else if(xch == 'E') { cMarkup = entityMarkup; curBLK = Xdef.BLK_ENTITY; }
				else if(xch == '[') { cMarkup = cdataMarkup; markCHAR = ']'; }
				else if(xch == 'D') { cMarkup = doctypeMarkup; curBLK = Xdef.BLK_DOCTYPE1; }
				else throwError(1); //Error: Invalid Char

				markupPos = 0;  markCount = cMarkup.length-1;
				do
				{	i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(markupPos > markCount) break;
					if(xch == cMarkup[markupPos]) markupPos++;
					else throwError(1);	//Error: Invaid Tag Char
				} while(true);
				continue CoreLoop;
			}


			else if(curBLK == Xdef.BLK_ENTITY)
			{
				do
				{	if(xch == '\"' || xch == '\'') { markCHAR = xch; break; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);

				iFlag = false; markCount = 0;
				do
				{	i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) {
						if(markCount > 0) break;
						if(xch == '>') markCount++;
						else if(xch == ' ' || xch == '	' || xch == '\r') ;
						else if(xch == '\n') { lineCount++; charPos = i; }
						else throwError(1);	//Error: Invalid Char
					}
					else if(xch == markCHAR)
					iFlag = true;
				} while(true);
				curBLK = Xdef.BLK_DOCTYPE2;
			}


			else if(curBLK == Xdef.BLK_DOCTYPE2)
			{
				iFlag = false; markCount = 0;
				do
				{	if(iFlag) {
						if(markCount > 0) break;
						if(xch == markCHAR) markCount++;
						else if(markCHAR == '>' && (xch == ' ' || xch == '	' || xch == '\r' || xch == '\n'))
							{ if(xch == '\n') { lineCount++; charPos = i; } }
						else throwError(1); //Error Invalid Char
					}
					else {
						//**** all the commented portion in this block is a temporary hack !
						//if(xch == '<') { markCHAR = '!';  curBLK = Xdef.BLK_MISC; iFlag = true; }
						//else if(xch == ' ' || xch == '	' || xch == '\r');
						if(xch == ' ' || xch == '	' || xch == '\r');
						else if(xch == '\n') { lineCount++; charPos = i; }
						else if(xch == ']') { markCHAR = '>';  curBLK = Xdef.BLK_TAGBEGIN; iFlag = true; }
						//else throwError(1); //Error: Invalid Char
					}
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
				} while(true);
			}


			else if(curBLK == Xdef.BLK_DOCTYPE1)
			{
				iFlag = false; markCount = 0;
				do
				{	if(xch == '[') { curBLK = Xdef.BLK_DOCTYPE2; iFlag = true; }
					else if(xch == '>') { curBLK = Xdef.BLK_TAGBEGIN; iFlag = true; }
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) break;
				} while(true);

			}


			else if(curBLK == Xdef.BLK_PI)
			{
				iFlag = false; markCount = 0;
				do
				{	if(markCount > 0 && xch == '>') iFlag = true;
					else if(xch == '?') markCount++;
					else markCount = 0;
					xdata[xdataLen] = xch; xdataLen++;
					i++;	if(i > charsRead) refreshxBuffer();
					xch = xbuf[i];
					if(iFlag) { curBLK = handleChrData(); tagOpen = false; break; }
				} while(true);
			}


			//generic loop for parsing whitespaces
			do
			{	if(xch == ' ' || xch == '	' || xch == '\r') ;
				else if(xch == '\n') { lineCount++; charPos = i; }
				else break;
				i++;
				if(i > charsRead) refreshxBuffer();
				xch = xbuf[i];
			} while(true);

		} while(true); //End of CoreLoop

		} while(true); //End of ParserLoop


		//--------------------------------------------------------------
		//                    end of parser logic
		//--------------------------------------------------------------
		} catch(IOException e) { System.out.println(e.toString()); Doc=null; return; }
		  catch(ArrayIndexOutOfBoundsException ei) { }
		  catch(InstantiationException ee) { Doc=null; return; }

		//Make sure root element is closed
		if(cNodeDepth >= 0)
		{	System.out.println("Parser Error: Document is not well-formed");
			Doc = null;  return;
		}

		//Do final settings
		Doc.nextNode.prevNode = null;
		Doc.nextNode = null;
	}


	void refreshxBuffer() throws IOException
	{
		charPos -= i-1;
		i = 0;

		do {
			charsRead = xstream.read(xbuf, 0, xBufSize) -1;
			if(charsRead == 0) continue;
			else if(charsRead < -1) xbuf = new char[0];
			return;
		} while(true);
	}


	short handleTAG() throws InstantiationException
	{
		String nStr = new String(xdata, 0, xdataLen);
//System.out.println("========== In Handle Tag "+tagQualifier+nStr+"===========");

		short newBlk = Xdef.BLK_TAGNAME;

		if(tagOpen) {
//System.out.println(" -----> tagOpen is True");
			attrName[attrCount] = nStr;
			newBlk = Xdef.BLK_ATTR;
		}
		else {
			if(tagQualifier == '/') {	//Node closed
//System.out.println("*****Before Closing: "+nStr);
//System.out.println("cNode: "+cNode.nodeChr+"  cNodeName: "+cNodeName);
				if(nStr.equals(cNodeName))
				{	if(attrCount > 0) addAttributes(cNode);
					sNode = cNode;
					cNode = cNode.parentNode;
					cNodeName = cNode.nodeChr;
					cNodeDepth--;
//System.out.println("*****After Closing: "+nStr);
//System.out.println("cNode: "+cNode.nodeChr+"  cNodeName: "+cNodeName);
				}
				else { System.out.println(nStr); throwError(3); }
				newBlk = Xdef.BLK_TAGEND;
			}
			else if(tagQualifier == '?') {
				newBlk = Xdef.BLK_PI;
			}
			else {
				tNode = createNode(Node.ELEMENT_NODE, nStr);
				if(cNodeDepth < 0) //document root
				{	if(cNodeDepth > -99) throwError(3); //Attempt to create more than one root element
					cNode = tNode;
					Doc.nodeData[Xdef.IDOCUMENT_ROOT] = cNode;
					cNode.nodeDepth = cNodeDepth = 0;
					cNode.parentNode = Doc;	//Not required, but set to prevent null-pointer-Ex
					eNode=Doc;	//to be reset at the end
				}
				else
				{	if(attrCount > 0) addAttributes(cNode);
					tNode.parentNode = cNode;
					cNode = tNode;
					cNodeDepth++;
					cNode.nodeDepth = cNodeDepth;
					cNode.prevSibling = sNode;
					if(sNode != null) { sNode.nextSibling = cNode; sNode=null; }
				}
				cNodeName = nStr;
				eNode.nextNode = cNode;
				cNode.prevNode = eNode;
				eNode = cNode;
//System.out.println("*****Node Created: "+nStr);
//System.out.println("cNode: "+cNode.nodeChr+"  cNodeName: "+cNodeName);
			}
			tagOpen = true;
		}

		xdataLen = 0;
		tagQualifier = ' ';
		return newBlk;
	}

	void closeTAG() throws InstantiationException
	{
		if(tagOpen)	//Close current Node
		{
//System.out.println("*****Before Closing: ");
//System.out.println("cNode: "+cNode.nodeChr+"  cNodeName: "+cNodeName);
			if(attrCount > 0) addAttributes(cNode);
			sNode = cNode;
			cNode = cNode.parentNode;
			cNodeName = cNode.nodeChr;
			cNodeDepth--;
//System.out.println("*****After Closing: ");
//System.out.println("cNode: "+cNode.nodeChr+"  cNodeName: "+cNodeName);
		}
		else throwError(1);	//Error: Invalid Char at i-1
		tagQualifier = ' ';
	}

	short handleAttr()
	{
		String nStr = new String(xdata, 0, xdataLen);
		attrValue[attrCount] = nStr;
		attrCount++;

		xdataLen = 0;
		equalCount = 0;
		return Xdef.BLK_TAGNAME;
	}

	short handleChrData()
	{
		if(xdataLen < 1) return Xdef.BLK_TAGBEGIN;
		String nStr = new String(xdata, 0, xdataLen);
//System.out.println("*****In HandleChrData. curBLK: "+curBLK+"  DataLen: "+xdataLen+"  Data:");
//System.out.println(nStr);
		if(attrCount > 0) addAttributes(cNode);
		if(curBLK == Xdef.BLK_TEXTDATA)
		{
			tNode = createNode(Node.TEXT_NODE, nStr);
			tNode.parentNode = cNode;
			tNode.nodeDepth = cNodeDepth+1;
			tNode.prevSibling = sNode;
			if(sNode != null) sNode.nextSibling = tNode;
			sNode = tNode;
			eNode.nextNode = tNode;
			tNode.prevNode = eNode;
			eNode = tNode;
		}
		xdataLen = 0;
		return Xdef.BLK_TAGBEGIN;
	}

	void throwError(int errNo) throws InstantiationException
	{	String errPos = " line "+lineCount+" char "+(i-charPos);

		switch(errNo)
		{
			case 1:
				System.out.println("Malformed Document Error: Invalid Character at"+errPos);
				break;

			case 2:
				System.out.println("Null Tag Name at"+errPos);
				break;

			case 3:
				System.out.println("Parse Error: Attempt to create more than one root element");
				break;

			default:
				System.out.println("Parse Error at"+errPos);
		}

		throw new InstantiationException();
	}

	//XML Utils
	XMLNode createNode(short nType, String data)
	{
		XMLNode xn  = new XMLNode();
		xn.nodeType = nType;
		xn.ownerDoc = Doc;
		xn.nodeChr  = data;
		//xn.nodeData = new Object[Xdef.ELEMENT_DATA_COUNT];
		return xn;
	}

	void addAttributes(XMLNode tN)
	{
		tN.nodeData = new Object[Xdef.ELEMENT_DATA_COUNT];
		tN.nodeData[Xdef.IELEMENT_ATTRCOUNT] = new int[] { attrCount };
		String[] attrN = new String[attrCount];
		String[] attrV = new String[attrCount];
		System.arraycopy(attrName,  0, attrN, 0, attrCount);
		System.arraycopy(attrValue, 0, attrV, 0, attrCount);
		tN.nodeData[Xdef.IELEMENT_ATTRNAME]  = attrN;
		tN.nodeData[Xdef.IELEMENT_ATTRVALUE] = attrV;
		attrCount = 0;
	}


	/*public static void main(String[] args)
	{
		XMLParser tp = new XMLParser();
		Document doc;
		if(args.length < 1 || args[0].equals("1"))
			//doc = tp.parseFile("design.xsl");
			doc = tp.parseFile("Test.xml");
		else if(args[0].equals("2"))
			doc = tp.parseFile("REC-xml-19980210.xml");
		else
			doc = tp.parseFile("allCDR_6.xml");

		//System.out.println("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
		//System.out.println("                          Document Tree");
		//System.out.println("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");

		Element root = doc.getDocumentElement();
		//System.out.println("Root Element: "+((XMLNode)(root)).nodeChr);
		//XMLNode tNode = ((XMLNode)root).nextNode;
		//for(;tNode != null; tNode=tNode.nextNode)
		//System.out.println(tNode.nodeID+" <"+tNode.nodeChr+">  prevSibling: "+((tNode.prevSibling==null) ? "null":String.valueOf(tNode.prevSibling.nodeID))+"  nextSibling: "+((tNode.nextSibling==null) ? "null":String.valueOf(tNode.nextSibling.nodeID)));

		//NodeList tList = root.getElementsByTagName("BODY");
		NodeList tList = root.getChildNodes();
		//tList = (tList.item(0)).getChildNodes();
		//tList = ((Element)tList.item(0)).getElementsByTagName("*");
		int len = tList.getLength();

		System.out.println("Total Elements: "+len);
		//System.out.println("*************** Root Element: "+((XMLNode)(root)).nodeChr);
		//tNode = ((XMLNode)root).nextNode;
		//tNode = ((XMLNode)tList).nextNode;
		//for(;tNode != null; tNode=tNode.nextSibling)
		//System.out.println(tNode.nodeID+" <"+tNode.nodeChr+">  prevSibling: "+((tNode.prevSibling==null) ? "null":String.valueOf(tNode.prevSibling.nodeID))+"  nextSibling: "+((tNode.nextSibling==null) ? "null":String.valueOf(tNode.nextSibling.nodeID)));

		for(int i=0; i < len; i++)
			//System.out.println("---<"+tList.item(i).getNodeName()+" "+((XMLNode)tList.item(i)).nodeID+">---");
			tList.item(i).getNodeName();
		System.out.println("=================== XML Output =====================");
		//System.out.println(tp.getXML());
		try {
		tp.saveXML("Test.xml");
		} catch(Exception e) { System.out.println(e.toString()); }
		System.exit(0);
	}
	*/
}
